# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['cpraa']

package_data = \
{'': ['*'],
 'cpraa': ['.git/*',
           '.git/hooks/*',
           '.git/info/*',
           '.git/logs/*',
           '.git/logs/refs/heads/*',
           '.git/logs/refs/remotes/origin/*',
           '.git/objects/03/*',
           '.git/objects/04/*',
           '.git/objects/0d/*',
           '.git/objects/12/*',
           '.git/objects/14/*',
           '.git/objects/17/*',
           '.git/objects/1a/*',
           '.git/objects/1b/*',
           '.git/objects/21/*',
           '.git/objects/2d/*',
           '.git/objects/2e/*',
           '.git/objects/35/*',
           '.git/objects/39/*',
           '.git/objects/3c/*',
           '.git/objects/48/*',
           '.git/objects/4a/*',
           '.git/objects/4d/*',
           '.git/objects/50/*',
           '.git/objects/52/*',
           '.git/objects/67/*',
           '.git/objects/6a/*',
           '.git/objects/6b/*',
           '.git/objects/76/*',
           '.git/objects/7b/*',
           '.git/objects/7c/*',
           '.git/objects/7f/*',
           '.git/objects/83/*',
           '.git/objects/8d/*',
           '.git/objects/91/*',
           '.git/objects/93/*',
           '.git/objects/95/*',
           '.git/objects/9e/*',
           '.git/objects/a5/*',
           '.git/objects/a8/*',
           '.git/objects/a9/*',
           '.git/objects/af/*',
           '.git/objects/b0/*',
           '.git/objects/b2/*',
           '.git/objects/b8/*',
           '.git/objects/bd/*',
           '.git/objects/c8/*',
           '.git/objects/c9/*',
           '.git/objects/cc/*',
           '.git/objects/cd/*',
           '.git/objects/d1/*',
           '.git/objects/d3/*',
           '.git/objects/d5/*',
           '.git/objects/d9/*',
           '.git/objects/e8/*',
           '.git/objects/eb/*',
           '.git/objects/f2/*',
           '.git/objects/f3/*',
           '.git/objects/f5/*',
           '.git/objects/fc/*',
           '.git/objects/fe/*',
           '.git/objects/pack/*',
           '.git/refs/heads/*',
           '.git/refs/remotes/origin/*',
           'AFs/*',
           'AFs/KR21/*']}

install_requires = \
['PySMT>=0.9.0,<0.10.0',
 'cvxopt>=1.2.5,<2.0.0',
 'matplotlib>=3.3.4,<4.0.0',
 'numpy>=1.20.0,<2.0.0',
 'pypoman>=0.5.4,<0.6.0',
 'scipy>=1.6.0,<2.0.0',
 'z3-solver>=4.8.8,<5.0.0']

setup_kwargs = {
    'name': 'cpraa',
    'version': '0.7.0',
    'description': 'A checker for probabilistic abstract argumentation',
    'long_description': None,
    'author': 'Nikolai Käfer',
    'author_email': 'nikolai.kaefer@tu-dresden.de',
    'maintainer': 'Nikolai Käfer',
    'maintainer_email': 'nikolai.kaefer@tu-dresden.de',
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
